#ifndef __LIBSL_H__
#define __LIBSL_H__

#include <math.h>
#include <string>
#include <vector>
#include <algorithm>
#include <stdarg.h>

#include <OpenImageIO/ustring.h>

using OpenImageIO::ustring;

using std::min;
using std::max;

#include "vector3.h"
#include "matrix4.h"
#include "color3.h"

#include "fmath.h"
#include "libmath.h"

#include "public.h"

namespace xrt
{
	class ShaderEnvironment;
};

namespace SL
{
// Mathematical Functions

	inline float radians( float degrees ) { return M_PI * degrees / 180.0; }
	inline float degrees( float radians ) { return 180.0 * radians / M_PI; }
	inline float pow( float x, float y )
	{
		if ( x <= 0.0f ) 
		{
			float yf = (float)floor (y);
			if ( yf == y && yf >= 1.0f ) 		     
				return ::pow( x, y );
			else 
				return 0.0f;
		}
		else
		{
			float ylnx = y * log( x );
			if ( ylnx <= -87.0f )
				return 0.0f;
			else 
				return exp( ylnx );
		}
	}

	inline float inversesqrt( float x ) { return 1 / sqrt( x ); }

	inline float mod( float x, float y ) 
	{ 
		if (y == 0)
			return x;
		else
		{
			float result = fmodf(x, y); 
			return result >= 0 ? result : result + y; 
		}
	}
	//inline int mod( int x, int y ) { int result = x % y; return result >= 0 ? result : result + y; }
	//inline float abs( float x ) { return x < 0 ? -x : x; }
	inline float sign( float x ) { return x < 0 ? -1 : x > 0 ? 1 : 0; }

	inline float min( float x, float y )
	{
		return ::min( x, y );
	}
	inline float max( float x, float y )
	{
		return ::max( x, y );
	}

	template <typename T> T min( const T &a, const T &b )
	{
		return T( ::min( a[0], b[0] ), ::min( a[1], b[1] ), ::min( a[2], b[2] ) );
	}
	template <typename T> T min( const T &a, float f )
	{
		return min( a, T(f) );
	}
	template <typename T> T min( float f, const T &a )
	{
		return min( a, T(f) );
	}
	template <typename T> T max( const T &a, const T &b )
	{
		return T( ::max( a[0], b[0] ), ::max( a[1], b[1] ), ::max( a[2], b[2] ) );
	}
	template <typename T> T max( const T &a, float f )
	{
		return max( a, T(f) );
	}
	template <typename T> T max( float f, const T &a )
	{
		return max( a, T(f) );
	}

	inline float min ( float x, float y, int n, ... )
	{
		va_list args;

		va_start( args, n );

		float m = ::min( x, y );

		for ( int i = 0; i < n; i++ )
			m = ::min( m, va_arg( args, double ) );

		return m;
	}

	inline float max ( float x, float y, int n, ... )
	{
		va_list args;

		va_start( args, n );

		float m = ::max( x, y );

		for ( int i = 0; i < n; i++ )
			m = ::max( m, va_arg( args, double ) );

		return m;
	}

	template <typename T> T min(const T &a, const T &b, int n, ... )
	{
		va_list args;

		va_start( args, n );

		T m = ::min( x, y );

		for ( int i = 0; i < n; i++ )
			m = ::min( m, va_arg( args, T ) );

		return m;
	}

	template <typename T> T max( const T &a, const T &b, int n, ... )
	{
		va_list args;

		va_start( args, n );

		T m = ::max( x, y );

		for ( int i = n - 2; i ; i-- )
			m = ::max( m, va_arg( args, T ) );

		return m;
	}

	inline float clamp( const float x, const float min, const float max ) { return x < min ? min : x > max ? max : x; }
	inline int clamp( const int x, const int min, const int max ) { return x < min ? min : x > max ? max : x; }

	template <typename T> T clamp( const T &x, const T &min, const T &max )
	{
		return T( clamp( x[0], min[0], max[0] ), clamp( x[1], min[1], max[1] ), clamp( x[2], min[2], max[2] ) );
	}

	inline float mix( float x, float y, float alpha )
	{
		return ( 1.0f - alpha ) * x + alpha * y;
	}

	template <typename T> T mix( const T &x, const T &y, const float alpha )
	{
		return ( 1.0f - alpha ) * x + alpha * y;
	}

	template <typename T> T mix( const T &x, const T &y, const T& alpha )
	{
		return ( 1.0f - alpha ) * x + alpha * y;
	}

	inline float round ( float x ) 
	{ 
		return x > 0 ? (float) ( (int) ( x + 0.5f ) ) : (float) ( (int) ( x - 0.5f ) ); 
	}

	inline float step( float min, float x ) { return x < min ? 0 : 1; }
	inline float smoothstep( float min, float max, float x ) 
	{ 
		if ( x > max )
			return 1;
		else if ( x < min )
			return 0;
		else
		{
			float val = ( x - min ) / ( max - min );
			return ( 3 - 2 * val ) * val * val;
		}
	}
	EXPORT float filterstep2( xrt::ShaderEnvironment&, float edge, float value, ... );
	EXPORT float filterstep3( float edge, float value1, float value2, ... );
	EXPORT float splinef ( float, unsigned int, ...);
	EXPORT float splinef ( ustring basis, float, unsigned int, ...);
	EXPORT float splinef ( float, std::vector<float>& );
	EXPORT float splinef ( ustring basis, float, std::vector<float>& );
	EXPORT xrt::Color3 splinec ( float, unsigned int, ...);
	EXPORT xrt::Color3 splinec ( ustring basis, float, unsigned int, ...);
	EXPORT xrt::Color3 splinec ( float, std::vector<xrt::Color3>& );
	EXPORT xrt::Color3 splinec ( ustring basis, float, std::vector<xrt::Color3>& );
	EXPORT xrt::Vector3 splinep ( float, unsigned int, ...);
	EXPORT xrt::Vector3 splinep ( ustring basis, float, unsigned int, ...);
	EXPORT xrt::Vector3 splinep ( float, std::vector<xrt::Vector3>& );
	EXPORT xrt::Vector3 splinep ( ustring basis, float, std::vector<xrt::Vector3>& );
	EXPORT float Du ( xrt::ShaderEnvironment&, float ), Dv ( xrt::ShaderEnvironment&, float ), Deriv ( xrt::ShaderEnvironment&, float, float );
	EXPORT xrt::Color3 Du ( xrt::ShaderEnvironment&, xrt::Color3& ), Dv ( xrt::ShaderEnvironment&, xrt::Color3& ), Deriv ( xrt::ShaderEnvironment&, xrt::Color3&, float );
	EXPORT xrt::Vector3 Du ( xrt::ShaderEnvironment&, xrt::Vector3& ), Dv ( xrt::ShaderEnvironment&, xrt::Vector3& ), Deriv ( xrt::ShaderEnvironment&, xrt::Vector3&, float );
	inline float randomf() { return random(); }
	inline xrt::Color3 randomc() { return xrt::Color3( random(), random(), random() ); }
	inline xrt::Vector3 randomp() { return xrt::Vector3( random(), random(), random() ); }
	EXPORT float noise ( float ), noise ( float, float ), 
		noise ( float, float, float ), noise ( float, float, float, float );
	inline float noisef ( float u ) { return 0.5f * noise( u ) + 0.5f; }
	inline float noisef ( float u, float v ) { return 0.5f * noise( u, v ) + 0.5f; }
	inline float noisef ( xrt::Vector3 &p ) { return 0.5f * noise( p.x, p.y, p.z ) + 0.5f; }
	inline float noisef ( xrt::Vector3 &p, float t ) { return 0.5f * noise( p.x, p.y, p.z, t ) + 0.5f; }
	inline xrt::Vector3 noisep ( float u ) 
	{ 
		return 0.5f * xrt::Vector3( noise( u ), noise( u + u ), noise( u + u + u ) ) + 0.5f; 
	}
	inline xrt::Vector3 noisep ( float u, float v ) 
	{ 
		return 0.5f * xrt::Vector3 ( noise( u, v ), noise( v, u ), noise( u + v, u - v ) ) + 0.5f; 
	}
	inline xrt::Vector3 noisep ( xrt::Vector3 &p ) 
	{ 
		return 0.5f * xrt::Vector3 ( noise( p.x, p.y, p.z ), noise( p.y, p.z, p.x ), noise( p.z, p.x, p.y ) ) + 0.5f; 
	}
	inline xrt::Vector3 noisep ( xrt::Vector3 &p, float t ) 
	{ 
		return 0.5f * xrt::Vector3 ( noise( p.x, p.y, p.z, t ), noise( p.y, p.z, p.x, t ), noise( p.z, p.x, p.y, t ) ) + 0.5f; 
	}
	inline xrt::Color3 noisec ( float u ) 
	{ 
		return 0.5f * xrt::Color3( noise( u ), noise( u + u ), noise( u + u + u ) ) + 0.5f; 
	}
	inline xrt::Color3 noisec ( float u, float v ) 
	{ 
		return 0.5f * xrt::Color3 ( noise( u, v ), noise( v, u ), noise( u + v, u - v ) ) + 0.5f; 
	}
	inline xrt::Color3 noisec ( xrt::Vector3 &p ) 
	{ 
		return 0.5f * xrt::Color3 ( noise( p.x, p.y, p.z ), noise( p.y, p.z, p.x ), noise( p.z, p.x, p.y ) ) + 0.5f; 
	}
	inline xrt::Color3 noisec ( xrt::Vector3 &p, float t ) 
	{ 
		return 0.5f * xrt::Color3 ( noise( p.x, p.y, p.z, t ), noise( p.y, p.z, p.x, t ), noise( p.z, p.x, p.y, t ) ) + 0.5f; 
	}
	xrt::Vector3 noisev ( float ), noisev ( float , float ), noisev ( xrt::Vector3& ), noisev ( xrt::Vector3&, float );

	inline float pnoisef ( float u, float uperiod )
	{
		return noisef( mod( u, uperiod ) );
	}
	inline float pnoisef ( float u, float v, float uperiod, float vperiod )
	{
		return noisef( mod( u, uperiod ), mod( v, vperiod ) );
	}
	inline float pnoisef ( xrt::Vector3& p, xrt::Vector3& period )
	{
		return 0.5f * noise( mod( p.x, period.x ), mod( p.y, period.y ), mod( p.z, period.z ) ) + 0.5f;
	}
	inline float pnoisef ( xrt::Vector3& p, float t, xrt::Vector3& period, float tperiod )
	{
		return 0.5f * noise( mod( p.x, period.x ), mod( p.y, period.y ), mod( p.z, period.z ), mod( t, tperiod ) ) + 0.5f;
	}
	xrt::Color3 pnoisec ( float, float  ), pnoisec ( float , float, float , float  ), pnoisec ( xrt::Vector3&, xrt::Vector3&  ), pnoisec ( xrt::Vector3&, float, xrt::Vector3&, float  );
	xrt::Vector3 pnoisep ( float, float  ), pnoisep ( float , float, float , float  ), pnoisep ( xrt::Vector3&, xrt::Vector3&  ), pnoisep ( xrt::Vector3&, float, xrt::Vector3&, float  );
	xrt::Vector3 pnoisev ( float, float  ), pnoisev ( float , float, float , float  ), pnoisev ( xrt::Vector3&, xrt::Vector3&  ), pnoisev ( xrt::Vector3&, float, xrt::Vector3&, float  );
	EXPORT float cellnoisef ( float u ); 
	EXPORT float cellnoisef ( float u, float v );
	EXPORT float cellnoisef ( float u, float v, float w );
	EXPORT float cellnoisef ( float u, float v, float w, float t );
	EXPORT float cellnoisef ( xrt::Vector3 &p );
	EXPORT float cellnoisef ( xrt::Vector3 &p, float t );
	EXPORT xrt::Vector3 cellnoisep ( float u ); 
	EXPORT xrt::Vector3 cellnoisep ( float u, float v ); 
	EXPORT xrt::Vector3 cellnoisep ( xrt::Vector3 &p ); 
	EXPORT xrt::Vector3 cellnoisep ( xrt::Vector3 &p, float t ); 
	EXPORT xrt::Color3 cellnoisec ( float u ); 
	EXPORT xrt::Color3 cellnoisec ( float u, float v ); 
	EXPORT xrt::Color3 cellnoisec ( xrt::Vector3 &p );
	EXPORT xrt::Color3 cellnoisec ( xrt::Vector3 &p, float t ); 
	xrt::Vector3 cellnoisev ( float ), cellnoisev ( float , float ), cellnoisev ( xrt::Vector3& ), cellnoisev ( xrt::Vector3&, float );

	// Geometric Functions

	inline float xcomp( const xrt::Vector3& p ) { return p.x; };
	inline float ycomp( const xrt::Vector3& p ) { return p.y; };
	inline float zcomp( const xrt::Vector3& p ) { return p.z; };
	inline void setxcomp( xrt::Vector3& p, const float f ) { p.x = f; };
	inline void setycomp( xrt::Vector3& p, const float f ) { p.y = f; };
	inline void setzcomp( xrt::Vector3& p, const float f ) { p.z = f; };
	inline float comp ( const xrt::Vector3 &p, float index ) { return p[(int)index]; }
	inline void setcomp ( xrt::Vector3 &p, float index, float value ) { p[(int)index] = value; }
	inline float length( const xrt::Vector3& v ) { return v.length(); }
	inline float distance( const xrt::Vector3& a, const xrt::Vector3& b ) { return SL::length( b - a ); }
	EXPORT float ptlined( const xrt::Vector3&, const xrt::Vector3&, const xrt::Vector3& );
	EXPORT xrt::Vector3 rotate( xrt::Vector3&, float, xrt::Vector3&, xrt::Vector3& );
	EXPORT float area( xrt::ShaderEnvironment& sv, xrt::Vector3& p);  
	inline xrt::Vector3 faceforward( xrt::Vector3& N, xrt::Vector3& I, xrt::Vector3& Nref) { return sign( - dot( I, Nref ) ) * N; }
	EXPORT inline xrt::Vector3 faceforward( xrt::ShaderEnvironment& sv, xrt::Vector3& N, xrt::Vector3& I );
	inline xrt::Vector3 reflect( xrt::Vector3& I, xrt::Vector3& N) { return I - 2 * dot( I, N ) * N; }
	inline xrt::Vector3 refract( xrt::Vector3& I, xrt::Vector3& N, float eta )
	{
		float IdotN = dot( I, N );
		float k = 1.f - eta * eta * ( 1.f - IdotN * IdotN );
		return k < 0.f ? xrt::Vector3(0.f) : eta * I - ( eta * IdotN + sqrt( k ) ) * N;
	}
	EXPORT void fresnel( xrt::Vector3&, xrt::Vector3&, float, float &, float& );
	EXPORT void fresnel( xrt::Vector3&, xrt::Vector3&, float, float &, float&, xrt::Vector3&, xrt::Vector3& );
	EXPORT xrt::Vector3 transform( xrt::ShaderEnvironment&, ustring, xrt::Vector3& );
	EXPORT xrt::Vector3 transform( xrt::ShaderEnvironment&, ustring, ustring, xrt::Vector3& );
	EXPORT xrt::Vector3 transform( xrt::Matrix4&, xrt::Vector3& );
	xrt::Vector3 transform( xrt::ShaderEnvironment&, ustring, xrt::Matrix4&, xrt::Vector3& );
	EXPORT xrt::Vector3 vtransform( xrt::ShaderEnvironment&, ustring, xrt::Vector3& );
	EXPORT xrt::Vector3 vtransform( xrt::ShaderEnvironment&, ustring, ustring, xrt::Vector3& );
	EXPORT xrt::Vector3 vtransform( xrt::Matrix4&, xrt::Vector3& );
	xrt::Vector3 vtransform( xrt::ShaderEnvironment&, ustring, xrt::Matrix4&, xrt::Vector3& );
	EXPORT xrt::Vector3 ntransform( xrt::ShaderEnvironment&, ustring, xrt::Vector3& );
	EXPORT xrt::Vector3 ntransform( xrt::ShaderEnvironment&, ustring, ustring, xrt::Vector3& );
	EXPORT xrt::Vector3 ntransform( xrt::Matrix4&, xrt::Vector3& );
	xrt::Vector3 ntransform( xrt::ShaderEnvironment&, ustring, xrt::Matrix4&, xrt::Vector3& );
	EXPORT float depth( xrt::ShaderEnvironment&, xrt::Vector3& );
	EXPORT xrt::Vector3 calculatenormal( xrt::ShaderEnvironment& sv, xrt::Vector3& p );

	// Color functions

	inline float comp ( xrt::Vector3 &c, float index ) { return c[(int)index]; }
	inline float comp ( xrt::Color3 &c, float index ) { return c[(int)index]; }
	inline void setcomp ( xrt::Color3 &c, float index, float value ) { c[(int)index] = value; }
	EXPORT xrt::Color3 ctransform ( ustring tospace, xrt::Color3& c );
	EXPORT xrt::Color3 ctransform ( ustring fromspace, ustring tospace, xrt::Color3 &c );
	inline float luminance( xrt::Color3 &c ) { return 0.114 * c.r + 0.587 * c.g + 0.299 * c.b; }

	// Matrix functions

	inline float comp( const xrt::Matrix4& m, float row, float column ) { return m[(int)row][(int)column]; }
	inline void setcomp( xrt::Matrix4& m, float row, float column, float value ) { m[(int)row][(int)column] = value; }
	inline float determinant( const xrt::Matrix4& m ) { return m.determinant(); }
	EXPORT xrt::Matrix4 translate ( const xrt::Matrix4& m, const xrt::Vector3& t );
	inline xrt::Matrix4 rotate ( const xrt::Matrix4& m, const float angle, const xrt::Vector3& axis ) { return m * xrt::Matrix4::Rotation (angle, axis); }
	EXPORT xrt::Matrix4 scale ( const xrt::Matrix4 &m, const xrt::Vector3& s );
	EXPORT xrt::Matrix4 mtransform( xrt::ShaderEnvironment&, ustring, xrt::Matrix4& );
	EXPORT xrt::Matrix4 mtransform( xrt::ShaderEnvironment&, ustring, ustring, xrt::Matrix4& );
	EXPORT xrt::Matrix4 mtransform( xrt::Matrix4&, xrt::Matrix4& );
	xrt::Matrix4 mtransform( xrt::ShaderEnvironment&, ustring, xrt::Matrix4&, xrt::Matrix4& );

	// Array functions
	
	template <typename T> float arraylength( T ) { return -1; }
	template <typename T> float arraylength( const std::vector<T> &array ) { return array.size(); }
		
	// String functions

	#if 0
	string concat ( string a, b, ... )
	#endif
	EXPORT void printf ( xrt::ShaderEnvironment&, ustring pattern, ... );
	EXPORT ustring format ( ustring pattern, ... );
	#if 0
	float match ( string pattern, subject )
	#endif

	// Shading and lighting functions

	void startIlluminance( xrt::ShaderEnvironment& );
	EXPORT void startIlluminance( xrt::ShaderEnvironment&, xrt::Vector3& );
	EXPORT void startIlluminance( xrt::ShaderEnvironment&, xrt::Vector3&, xrt::Vector3&, float );
	bool illuminance( xrt::ShaderEnvironment&, ustring , xrt::Vector3& );
	EXPORT bool illuminance0( xrt::ShaderEnvironment& );
	EXPORT bool illuminance1( xrt::ShaderEnvironment& );
	EXPORT bool illuminance3( xrt::ShaderEnvironment& );
	bool illuminance( xrt::ShaderEnvironment&, ustring, xrt::Vector3&, xrt::Vector3&, float );
	EXPORT bool illuminate( xrt::ShaderEnvironment&, xrt::Vector3& );
	EXPORT bool illuminate( xrt::ShaderEnvironment&, xrt::Vector3&, xrt::Vector3&, float );
	EXPORT bool solar( xrt::ShaderEnvironment& );
	EXPORT bool solar( xrt::ShaderEnvironment&, xrt::Vector3&, float );

	EXPORT xrt::Color3 ambient( xrt::ShaderEnvironment& );
	EXPORT xrt::Color3 diffuse( xrt::ShaderEnvironment&, xrt::Vector3& );
	EXPORT xrt::Color3 specular( xrt::ShaderEnvironment&, xrt::Vector3&, xrt::Vector3&, float );
	EXPORT xrt::Color3 specularbrdf ( xrt::ShaderEnvironment&, xrt::Vector3&,  xrt::Vector3&, xrt::Vector3&, float );
	EXPORT xrt::Color3 phong( xrt::ShaderEnvironment&, xrt::Vector3&, xrt::Vector3&, float );
	EXPORT float tracef( xrt::ShaderEnvironment&, xrt::Vector3&, xrt::Vector3&, ... );
	EXPORT xrt::Color3 tracec( xrt::ShaderEnvironment&, xrt::Vector3&, xrt::Vector3&, int ... );
	EXPORT xrt::Color3 transmission( xrt::ShaderEnvironment&, const xrt::Vector3&, const xrt::Vector3&, int, ... );
	EXPORT float occlusion3( xrt::ShaderEnvironment&, const xrt::Vector3&, const xrt::Vector3&, float, int, ... );
	EXPORT float occlusion2( xrt::ShaderEnvironment&, const xrt::Vector3&, const xrt::Vector3&, int, ... );
	EXPORT xrt::Color3 indirectdiffuse3( xrt::ShaderEnvironment&, const xrt::Vector3&, const xrt::Vector3&, float, int, ... );
	EXPORT xrt::Color3 indirectdiffuse2( xrt::ShaderEnvironment&, const xrt::Vector3&, const xrt::Vector3&, int, ... );

	// Texture mapping functions

	EXPORT float texturef ( xrt::ShaderEnvironment&, ustring, int, int, ... );
	EXPORT float texture2f ( xrt::ShaderEnvironment&, ustring, int, float, float, int, ... );
	EXPORT float texture8f ( xrt::ShaderEnvironment&, ustring, int, float, float, float, float, float, float, float, float, int, ... );
	EXPORT xrt::Color3 texturec ( xrt::ShaderEnvironment&, ustring, int, int, ... );
	EXPORT xrt::Color3 texture2c ( xrt::ShaderEnvironment&, ustring, int, float, float, int, ... );
	EXPORT xrt::Color3 texture8c ( xrt::ShaderEnvironment&, ustring, int, float, float, float, float, float, float, float, float, int, ... );
	float environmentf ( xrt::ShaderEnvironment&, ustring, int, xrt::Vector3 &, int, ... );
	float environmentf ( xrt::ShaderEnvironment&, ustring, int, xrt::Vector3 &, xrt::Vector3 &, xrt::Vector3 &, xrt::Vector3 &, int, ... );
	EXPORT xrt::Color3 environmentc ( xrt::ShaderEnvironment&, ustring, int, xrt::Vector3 &, int, ... );
	EXPORT xrt::Color3 environmentc ( xrt::ShaderEnvironment&, ustring, int, xrt::Vector3 &, xrt::Vector3 &, xrt::Vector3 &, xrt::Vector3 &, int, ... );
	EXPORT float shadowf( xrt::ShaderEnvironment&, ustring, int, xrt::Vector3 &, int, ... );
	EXPORT xrt::Color3 shadowc( xrt::ShaderEnvironment&, ustring, int, xrt::Vector3 &, int, ... );
	//float shadow( xrt::ShaderEnvironment&, const char*, int, xrt::Vector3 &, xrt::Vector3 &, xrt::Vector3 &, xrt::Vector3 &, ... );
	EXPORT float textureinfo ( xrt::ShaderEnvironment&, ustring, ustring, float& );
	EXPORT float textureinfo ( xrt::ShaderEnvironment&, ustring, ustring, std::vector<float>& );
	float textureinfo ( ustring, ustring, xrt::Color3& );
	float textureinfo ( ustring, ustring, xrt::Vector3& );
	float textureinfo ( ustring, ustring, xrt::Matrix4& );
	float textureinfo ( ustring, ustring, ustring& );

	// Message passing and information functions
	EXPORT float surface( xrt::ShaderEnvironment&, ustring& paramname, float &variable );
	EXPORT float surface( xrt::ShaderEnvironment&, ustring& paramname, xrt::Color3 &variable );
	EXPORT float surface( xrt::ShaderEnvironment&, ustring& paramname, xrt::Vector3 &variable );
	EXPORT float lightsource( xrt::ShaderEnvironment&, ustring& paramname, float &variable );
	EXPORT float lightsource( xrt::ShaderEnvironment&, ustring& paramname, xrt::Color3 &variable );
	EXPORT float lightsource( xrt::ShaderEnvironment&, ustring& paramname, xrt::Vector3 &variable );
	EXPORT float displacement( xrt::ShaderEnvironment&, ustring& paramname, float &variable );
	EXPORT float displacement( xrt::ShaderEnvironment&, ustring& paramname, xrt::Color3 &variable );
	EXPORT float displacement( xrt::ShaderEnvironment&, ustring& paramname, xrt::Vector3 &variable );
	EXPORT float volume( xrt::ShaderEnvironment&, ustring& paramname, float &variable );
	EXPORT float volume( xrt::ShaderEnvironment&, ustring& paramname, xrt::Color3 &variable );
	EXPORT float volume( xrt::ShaderEnvironment&, ustring& paramname, xrt::Vector3 &variable );
	#if 0
	float atmosphere( string paramname; output type variable )
	float incident( string paramname; output type variable )
	float opposite( string paramname; output type variable )
	string shadername ( )
	string shadername ( string shadertype )
	#endif
	EXPORT float attribute( xrt::ShaderEnvironment&, ustring&, float& );
	EXPORT float attribute( xrt::ShaderEnvironment&, ustring&, xrt::Vector3& );
	EXPORT float attribute( xrt::ShaderEnvironment&, ustring&, ustring& );
	EXPORT float option( xrt::ShaderEnvironment&, ustring&, float& );
	EXPORT float option( xrt::ShaderEnvironment&, ustring&, xrt::Vector3& );
	EXPORT float option( xrt::ShaderEnvironment&, ustring&, ustring& );
	EXPORT float rayinfo( xrt::ShaderEnvironment&, ustring&, float& );
	EXPORT float rayinfo( xrt::ShaderEnvironment&, ustring&, xrt::Vector3& );
	EXPORT float rayinfo( xrt::ShaderEnvironment&, ustring&, ustring& );
};

#endif //__LIBSL_H__
