#ifndef __SHADER_H__
#define __SHADER_H__

#include <stdarg.h>
#include <vector>
#include <map>
#include <string>

#include "refcount.h"
#include "paramtype.h"
#include "interpolator.h"
#include "transform.h"

#include "public.h"

namespace xrt
{
	class ParameterSet;
	class Parameter;
	class ShaderEnvironment;
	class Shape;

	class PUBLIC Shader: public ReferenceCounted
	{
	public:
		typedef void OpaqueParameters;

		struct Argument
		{
			Argument() {}
			Argument( const char* name, ParamType type, bool isOutput, size_t offset ):
				name( name ), type( type ), isOutput(isOutput), offset( offset ) {}
			ustring name;
			ParamType type;
			bool isOutput;
			size_t offset;
			void execute (ShaderEnvironment&, const Shape*, OpaqueParameters*) const;
			template <typename T> void execute (const Parameter& parameter, ShaderEnvironment& sv, const Shape* shape, OpaqueParameters* parameters) const
			{
				T *ptr;
				if (type.isarray())
				{
					((std::vector<T>*)((char*)parameters+offset))->resize(type.arraylen);
					ptr = ((std::vector<T>*)((char*)parameters+offset))->data();
				}
				else
					ptr = (T*)((char*)parameters+offset);

				shape->at (sv, parameter, ptr);
			}
		};

		struct ParameterInformation
		{
			const char *name; // name
			ParamType type; // data type
			std::vector<float> floats; // default float values
			std::vector<ustring> strings; // space name for matrices and
			// triples, for each array elem or default string values
		};

		struct ConnectionInfo
		{
			ParamType type;
			unsigned int id; // global parameter id; O if a shader parameter
			int index; // index in the array or tuple; -1 if no index
			unsigned int shader; // index of the shader in the shader group
			size_t offset; // offset in the parameter set
		};

		class ShaderConnectionInterface
		{
		public:
			ShaderConnectionInterface(const ConnectionInfo& source, const ConnectionInfo& destination) : source(source), destination(destination) {}
			void execute( ShaderEnvironment& sv, std::vector<OpaqueParameters*>& groupParameters, OpaqueParameters* out)
			{
				execute(sv, groupParameters.at(source.shader), out);
			}
			virtual void execute( ShaderEnvironment&, OpaqueParameters*, OpaqueParameters* ) = 0;
		protected:
			ConnectionInfo source, destination;
		};

		class PUBLIC ArgumentConnectionInterface
		{
		public:
			ArgumentConnectionInterface(const Parameter*, const Argument*);
			virtual void execute( ShaderEnvironment&, OpaqueParameters* ) = 0;
			virtual void createInterpolators( std::map< ustring, Transform >& ) = 0;
			virtual void deleteInterpolators() = 0;
		protected:
			const Argument *argument;
			ParamType type; // data type
			std::vector<float> floats; // default float values
			std::vector<ustring> strings; // space name for matrices and
			// triples, for each array elem or default string values
		};
		
		template< class T > class ArrayConnection: public ShaderConnectionInterface
		{
		public:
			ArrayConnection(const ConnectionInfo& source, const ConnectionInfo& destination) : ShaderConnectionInterface(source, destination) {}
			void execute( ShaderEnvironment&, OpaqueParameters* in, OpaqueParameters* out)
			{
				T *src, *dst;
				src = ((std::vector<T>*)((char*)in+source.offset))->data();
				dst = ((std::vector<T>*)((char*)out+destination.offset))->data();
				memcpy(dst, src, source.type.arraylen * sizeof(T));
			}
		};

		template< class T> static T* resolveAddress(ShaderEnvironment&, const ConnectionInfo&, OpaqueParameters*);

		class FloatConnection: public ShaderConnectionInterface
		{
		public:
			FloatConnection(const ConnectionInfo& source, const ConnectionInfo& destination) : ShaderConnectionInterface(source, destination) {}
			void execute(ShaderEnvironment&, OpaqueParameters*, OpaqueParameters*);
		};
			
		class TupleConnection: public ShaderConnectionInterface
		{
		public:
			TupleConnection(const ConnectionInfo& source, const ConnectionInfo& destination) : ShaderConnectionInterface(source, destination) {}
			void execute(ShaderEnvironment&, OpaqueParameters*, OpaqueParameters*);
		};

		class StringConnection: public ShaderConnectionInterface
		{
		public:
			StringConnection(const ConnectionInfo& source, const ConnectionInfo& destination) : ShaderConnectionInterface(source, destination) {}
			void execute(ShaderEnvironment&, OpaqueParameters*, OpaqueParameters*);
		};

		class MatrixConnection: public ShaderConnectionInterface
		{
		public:
			MatrixConnection(const ConnectionInfo& source, const ConnectionInfo& destination) : ShaderConnectionInterface(source, destination) {}
			void execute(ShaderEnvironment&, OpaqueParameters*, OpaqueParameters*);
		};

		class FloatTupleConnection: public ShaderConnectionInterface
		{
		public:
			FloatTupleConnection(const ConnectionInfo& source, const ConnectionInfo& destination) : ShaderConnectionInterface(source, destination) {}
			void execute(ShaderEnvironment&, OpaqueParameters*, OpaqueParameters*);
		};

	protected:
		struct Information
		{
			const char *name;
			unsigned int usedShaderEnvironment;
			unsigned int type;
		};

	public:
		template< class T > class ArgumentConnection: public ArgumentConnectionInterface
		{
		public:
			ArgumentConnection( const Shader*, const Parameter*, const Argument* );
			~ArgumentConnection();
			void execute( ShaderEnvironment&, OpaqueParameters* );
			void createInterpolators( std::map< ustring, Transform >& );

		private:
			std::vector< Interpolator< T > * > interpolators;
			void deleteInterpolators();
		};

		Shader();
		~Shader();
		virtual void evaluate( ShaderEnvironment&, void* ) = 0;
		virtual void init( ShaderEnvironment&, void* ) = 0;
		virtual const Information& getInformation() const = 0;
		virtual const Argument* getArguments() const = 0;
		virtual unsigned int getNumArguments() const = 0;
		virtual OpaqueParameters* createParameters() const = 0;
		virtual void releaseParameters(OpaqueParameters*) const = 0;
		void init( ShaderEnvironment&, const Shape*, void* );
		void init( ShaderEnvironment&, std::vector<void*>&, const Shape*, void* );
		const char* getLayerName() const { return layerName.c_str(); }
		void setLayerName( const char* layer ) { layerName = layer; }
		std::vector < ShaderConnectionInterface * >& connections() { return m_connections; }
		std::vector < ArgumentConnectionInterface * >& arguments() { return m_arguments; }
		void setCoordinateSystem( Transform &transform ) { shaderToWorld = transform; }
		const Transform& getCoordinateSystem() const { return shaderToWorld; }
		const Argument* find( ustring& ) const;
		const char * name() const { return getInformation().name; }
		unsigned int type() const { return getInformation().type; }
		unsigned int usedShaderEnvironment() const { return getInformation().usedShaderEnvironment; }
		bool getMessage( ustring &name, void* parameters, float& variable );
		bool getMessage( ustring &name, void* parameters, Color3& variable );
		bool getMessage( ustring &name, void* parameters, Vector3& variable );
	private:
		std::string layerName;
		std::vector < ShaderConnectionInterface* > m_connections;
		std::vector < ArgumentConnectionInterface* > m_arguments;
		Transform shaderToWorld;
	};

	template<> 
	EXPORT void Shader::ArgumentConnection<float>::createInterpolators( std::map< ustring, Transform >&  );
	template<> 
	EXPORT void Shader::ArgumentConnection<ustring>::createInterpolators( std::map< ustring, Transform >&  );
	template<> 
	EXPORT void Shader::ArgumentConnection<Color3>::createInterpolators( std::map< ustring, Transform >&  );
	template<> 
	EXPORT void Shader::ArgumentConnection<Vector3>::createInterpolators( std::map< ustring, Transform >&  );
	template<> 
	EXPORT void Shader::ArgumentConnection<Matrix4>::createInterpolators( std::map< ustring, Transform >&  );

	template< class T > 
	void Shader::ArgumentConnection< T >::deleteInterpolators()
	{
		unsigned int count = interpolators.size();

		for ( unsigned int i = 0; i < count; i++ )
			delete interpolators[i];

		interpolators.clear();
	}

	template< class T > 
	Shader::ArgumentConnection< T >::~ArgumentConnection()
	{
		deleteInterpolators();
	}

	EXPORT Shader *createShader( const std::string &, const std::string &, std::map< ustring, Transform > &, ParameterSet & );

};
#endif //__SHADER_H__
