// ======================================================================== //
// Copyright 2009-2012 Intel Corporation                                    //
//                                                                          //
// Licensed under the Apache License, Version 2.0 (the "License");          //
// you may not use this file except in compliance with the License.         //
// You may obtain a copy of the License at                                  //
//                                                                          //
//     http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                          //
// Unless required by applicable law or agreed to in writing, software      //
// distributed under the License is distributed on an "AS IS" BASIS,        //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. //
// See the License for the specific language governing permissions and      //
// limitations under the License.                                           //
// ======================================================================== //

#ifndef __TASKSCHEDULER_H__
#define __TASKSCHEDULER_H__

#include "sys/thread.h"
#include "sys/atomic.h"
#include "sys/mutex.h"
#include "sys/condition.h"

#include <vector>
#include <list>

namespace sys
{
  /*! Task scheduler implementing a stack of tasks. */
  class __public TaskScheduler
  {
  public:
    /*! Task queues */
    enum QUEUE { GLOBAL_FRONT, GLOBAL_BACK };

    /*! thread and core ID struct */
    struct ThreadInfo 
    {
      ThreadInfo () : id(0) {}
      ThreadInfo (size_t id) : id(id) {}
      size_t id;  // thread ID from 0 to getNumberOfLogicalThreads()-1
    };
    
    /*! the run function executed for each work item of the task */
    typedef void (*runFunction)(const ThreadInfo& thread, void* ptr, size_t elt);
    
    /*! complete function executed at the end of the task */
    typedef void (*completeFunction)(const ThreadInfo& thread, void* ptr);

  private:

	/* initialization structure for threads */
    struct Thread {
      Thread (size_t tid, TaskScheduler* scheduler) : tid(tid), scheduler(scheduler) {}
      size_t tid;
      TaskScheduler* scheduler;
    };

    /* task class */
    class Task
    {
    public:
      __forceinline Task() 
        : run(NULL), runData(NULL), complete(NULL), completeData(NULL), name(NULL) {}

      __forceinline Task(runFunction run, void* runData, size_t elts, completeFunction complete, void* completeData, const char* name)
        : run(run), runData(runData), complete(complete), completeData(completeData), started(elts), completed(elts), name(name) {}
      
    public:
      runFunction run;             //!< run function
      void* runData;               //!< data pointer to execute run function
      completeFunction complete;   //!< complete function
      void* completeData;          //!< data pointer to execute complete function
      Atomic started;              //!< counts the number of started task set elements
      Atomic completed;            //!< counts the number of completed task set elements
      const char* name;            //!< name of this task
    };

  public:

    /*! task scheduler constructor */
    TaskScheduler (size_t numThreads = 0);

    /*! destructor */
    ~TaskScheduler();

    /*! returns number of threads created */
    size_t getNumThreads() const { return numThreads; }
    
    /*! starts the tasking system */
    void start();

    /*! adds a task */
    void addTask(const ThreadInfo& thread, QUEUE queue,
                 runFunction run, void* runData, size_t elts = 1, 
                 completeFunction complete = NULL, void* completeData = NULL, 
                 const char* name = NULL);

     /*! stops the tasking system and waits for all tasks to finish */
    void stop();

	/*! house keeping */
    static void init(int numThreads = -1);
	
	/*! access to the task scheduler */
	static TaskScheduler* getInstance();
  private:

    /*! thread function */
    void run(size_t tid);
    /*! thread function */
    static void threadFunction(Thread* thread);
    
	/*! house keeping */
    static void cleanup();

    /* thread handling */
  private:
    volatile bool terminateThreads;
    std::vector<thread_t> threads;
    size_t numThreads;

  public:
    Mutex mainMutex;           //!< the task system get locked while used

    /* stack of tasks */
  private:
    Mutex taskMutex;           //!< protection of the task stack
    ConditionVariable taskCondition;   //!< condition to signal if data is in task stack
    std::list<Task*> tasks;

    /* signaling finished */
  private:
    Mutex finishMutex; 
    ConditionVariable finishCondition;
    Atomic activeTasks;
	/* singleton */
  private:
	static TaskScheduler* instance;

  };
}

#endif

